@@ -2,10 +2,13 @@ use rusoto_sqs::Message as SqsMessage;
 use tokio::sync::mpsc::{channel, Receiver, Sender};
 use log::*;
 
-use crate::event_handler::EventHandler;
+use crate::event_handler::{EventHandler, OutputEvent, Completion};
 use crate::event_retriever::PayloadRetriever;
 use crate::sqs_completion_handler::CompletionHandler;
 use crate::sqs_consumer::Consumer;
+use std::fmt::Debug;
+use crate::cache::Cache;
+use std::marker::PhantomData;
 
 #[derive(Copy, Clone, Debug)]
 pub enum ProcessorState {
@@ -15,7 +18,7 @@ pub enum ProcessorState {
 }
 
 #[derive(Clone)]
-pub struct EventProcessor<C, EH, Input, Output, ER, CH>
+pub struct EventProcessor<C, EH, Input, Output, ER, CH, CacheT, E>
 where
     C: Consumer + Clone + Send + Sync + 'static,
     EH: EventHandler<InputEvent = Input, OutputEvent = Output> + Send + Sync + Clone + 'static,
@@ -23,16 +26,20 @@ where
     Output: Send + Sync + Clone + 'static,
     ER: PayloadRetriever<Input> + Send + Sync + Clone + 'static,
     CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync + Clone + 'static,
+    CacheT: Cache<E> + Send + Sync + Clone + 'static,
+    E: Debug + Clone + Send + Sync + 'static,
 {
     consumer: C,
     completion_handler: CH,
     event_retriever: ER,
     event_handler: EH,
     state: ProcessorState,
+    cache: CacheT,
     self_actor: Option<EventProcessorActor>,
+    _p: PhantomData<E>
 }
 
-impl<C, EH, Input, Output, ER, CH> EventProcessor<C, EH, Input, Output, ER, CH>
+impl<C, EH, Input, Output, ER, CH, CacheT, E> EventProcessor<C, EH, Input, Output, ER, CH, CacheT, E>
 where
     C: Consumer + Clone + Send + Sync + 'static,
     EH: EventHandler<InputEvent = Input, OutputEvent = Output> + Send + Sync + Clone + 'static,
@@ -40,12 +47,15 @@ where
     Output: Send + Sync + Clone + 'static,
     ER: PayloadRetriever<Input> + Send + Sync + Clone + 'static,
     CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync + Clone + 'static,
+    CacheT: Cache<E> + Send + Sync + Clone + 'static,
+    E: Debug + Clone + Send + Sync + 'static,
 {
     pub fn new(
         consumer: C,
         completion_handler: CH,
         event_handler: EH,
         event_retriever: ER,
+        cache: CacheT,
     ) -> Self {
         Self {
             consumer,
@@ -53,19 +63,23 @@ where
             event_handler,
             event_retriever,
             state: ProcessorState::Waiting,
+            cache,
             self_actor: None,
+            _p: PhantomData,
         }
     }
 }
 
-impl<C, EH, Input, Output, ER, CH> EventProcessor<C, EH, Input, Output, ER, CH>
+impl<C, EH, Input, Output, ER, CH, CacheT, E> EventProcessor<C, EH, Input, Output, ER, CH, CacheT, E>
 where
     C: Consumer + Clone + Send + Sync + 'static,
     EH: EventHandler<InputEvent = Input, OutputEvent = Output> + Send + Sync + Clone + 'static,
     Input: Send + Clone + 'static,
     Output: Send + Sync + Clone + 'static,
     ER: PayloadRetriever<Input> + Send + Sync + Clone + 'static,
-CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync + Clone + 'static,
+    CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync + Clone + 'static,
+    CacheT: Cache<E> + Send + Sync + Clone + 'static,
+    E: Debug + Clone + Send + Sync + 'static,
 {
     pub async fn process_event(&mut self, event: SqsMessage) {
         // TODO: Handle errors
@@ -79,10 +93,19 @@ CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync +
                 // TODO: We could reset the message visibility to 0 so it gets picked up again?
             }
         };
+
         info!("Handling event");
-        let completed = match self.event_handler.handle_event(retrieved_event).await {
-            Ok(completed) => completed,
-            Err(e) => {
+        let output_event = self.event_handler.handle_event(retrieved_event).await;
+
+        match output_event.completed_event {
+            Completion::Total(completed) => {
+                info!("Marking all events complete - total success");
+                self.completion_handler.mark_complete(event, completed).await;
+            },
+            Completion::Partial((_completed, err)) => {
+                warn!("EventHandler was only partially successful: {:?}", err);
+            },
+            Completion::Error(e) => {
                 info!("Event handler failed: {:?}", e);
                 return
                 // TODO: Retry
@@ -90,8 +113,13 @@ CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync +
             }
         };
 
-        info!("Marking event complete");
-        self.completion_handler.mark_complete(event, completed).await;
+        info!("Caching");
+
+        for identity in output_event.identities {
+            if let Err(e) = self.cache.store(identity).await {
+                warn!("Failed to cache with: {:?}", e);
+            }
+        }
 
         if let ProcessorState::Started = self.state {
             self.consumer
@@ -119,7 +147,7 @@ pub enum EventProcessorMessage {
     stop_processing {},
 }
 
-impl<C, EH, Input, Output, ER, CH> EventProcessor<C, EH, Input, Output, ER, CH>
+impl<C, EH, Input, Output, ER, CH, CacheT, E> EventProcessor<C, EH, Input, Output, ER, CH, CacheT, E>
 where
     C: Consumer + Clone + Send + Sync + 'static,
     EH: EventHandler<InputEvent = Input, OutputEvent = Output> + Send + Sync + Clone + 'static,
@@ -127,6 +155,8 @@ where
     Output: Send + Sync + Clone + 'static,
     ER: PayloadRetriever<Input> + Send + Sync + Clone + 'static,
     CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync + Clone + 'static,
+    CacheT: Cache<E> + Send + Sync + Clone + 'static,
+    E: Debug + Clone + Send + Sync + 'static,
 {
     pub async fn route_message(&mut self, msg: EventProcessorMessage) {
         match msg {
@@ -143,7 +173,7 @@ pub struct EventProcessorActor {
 }
 
 impl EventProcessorActor {
-    pub fn new<C, EH, Input, Output, ER, CH>(mut actor_impl: EventProcessor<C, EH, Input, Output, ER, CH>) -> Self
+    pub fn new<C, EH, Input, Output, ER, CH, CacheT, E>(mut actor_impl: EventProcessor<C, EH, Input, Output, ER, CH, CacheT, E>) -> Self
     where
         C: Consumer + Clone + Send + Sync + 'static,
         EH: EventHandler<InputEvent = Input, OutputEvent = Output> + Send + Sync + Clone + 'static,
@@ -151,6 +181,8 @@ impl EventProcessorActor {
         Output: Send + Sync + Clone + 'static,
         ER: PayloadRetriever<Input> + Send + Sync + Clone + 'static,
         CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync + Clone + 'static,
+        CacheT: Cache<E> + Send + Sync + Clone + 'static,
+        E: Debug + Clone + Send + Sync + 'static,
     {
         let (sender, receiver) = channel(1);
 
@@ -191,7 +223,7 @@ impl EventProcessorActor {
     }
 }
 
-pub struct EventProcessorRouter<C, EH, Input, Output, ER, CH>
+pub struct EventProcessorRouter<C, EH, Input, Output, ER, CH, CacheT, E>
 where
     C: Consumer + Clone + Send + Sync + 'static,
     EH: EventHandler<InputEvent = Input, OutputEvent = Output> + Send + Sync + Clone + 'static,
@@ -199,13 +231,15 @@ where
     Output: Send + Sync + Clone + 'static,
     ER: PayloadRetriever<Input> + Send + Sync + Clone + 'static,
     CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync + Clone + 'static,
+    CacheT: Cache<E> + Send + Sync + Clone + 'static,
+    E: Debug + Clone + Send + Sync + 'static,
 {
     receiver: Receiver<EventProcessorMessage>,
-    actor_impl: EventProcessor<C, EH, Input, Output, ER, CH>,
+    actor_impl: EventProcessor<C, EH, Input, Output, ER, CH, CacheT, E>,
 }
 
 
-async fn route_wrapper<C, EH, Input, Output, ER, CH>(mut router: EventProcessorRouter<C, EH, Input, Output, ER, CH>)
+async fn route_wrapper<C, EH, Input, Output, ER, CH, CacheT, E>(mut router: EventProcessorRouter<C, EH, Input, Output, ER, CH, CacheT, E>)
 where
     C: Consumer + Clone + Send + Sync + 'static,
     EH: EventHandler<InputEvent = Input, OutputEvent = Output> + Send + Sync + Clone + 'static,
@@ -213,6 +247,8 @@ where
     Output: Send + Sync + Clone + 'static,
     ER: PayloadRetriever<Input> + Send + Sync + Clone + 'static,
     CH: CompletionHandler<Message=SqsMessage, CompletedEvent=Output> + Send + Sync + Clone + 'static,
+    CacheT: Cache<E> + Send + Sync + Clone + 'static,
+    E: Debug + Clone + Send + Sync + 'static,
 {
     while let Some(msg) = router.receiver.recv().await {
         router.actor_impl.route_message(msg).await;
